#lang racket
(require "const.rkt"
         "msg.rkt"
         "task.rkt")

;; A team is either
;;  #\r
;;  #\b

;; A player is
;;  (player team num num ball-or-#f task)
(define-struct player (team x y ball task)
  #:mutable)

;; A ball is
;;  (ball num num player)
(define-struct ball (x y carried-by)
  #:mutable)

(define field (for/vector ([y (in-range VSIZE)])
                (for/vector ([x (in-range HSIZE)])
                  #f)))
(define (field-ref x y)
  (vector-ref (vector-ref field y) x))
(define (field-set! x y p)
  (vector-set! (vector-ref field y) x p))

;; Create the players:
(define red-players (for/list ([i (in-range NUM-PLAYERS)])
                      (player #\r i 0 #f (start-task))))
(define blue-players (for/list ([i (in-range NUM-PLAYERS)])
                       (player #\b i (sub1 VSIZE) #f (start-task))))
(define players (append red-players blue-players))

;; Create the balls:
(unless (= 2 NUM-BALLS) (error "expected 2"))
(define balls (let ([ball-at 
                     (lambda (p)
                       (let ([b (ball (player-x p) (player-y p) p)])
                         (set-player-ball! p b)
                         b))])
                (vector (ball-at (list-ref red-players 0))
                        (ball-at (list-ref blue-players 0)))))

;; Put the players on the field:
(for ([p players])
  (field-set! (player-x p) (player-y p) p))

(define moves 0)
(define red-passes 0)
(define red-steals 0)
(define blue-passes 0)
(define blue-steals 0)

(define (show-field)
  (printf "\nMoves: ~a  Red passes+steals: ~a+~a  Blue passes+steals: ~a+~a\n"
          moves
          red-passes red-steals
          blue-passes blue-steals)
  (for ([i (in-range VSIZE)])
    (for ([j (in-range HSIZE)])
      (let ([p (field-ref j i)])
        (cond
          [(not p) (printf "_")]
          [else 
           (printf "~a"
                   (if (player-ball p)
                       (char-upcase (player-team p))
                       (player-team p)))
           ;; Sanity check:
           (when (player-ball p)
             (unless (eq? p (ball-carried-by (player-ball p)))
               (error "player has ball, but ball doesn't refer to player")))])))
    (newline))
  ;; More sanity checks:
  (for ([p players])
    (unless (eq? (field-ref (player-x p) (player-y p)) p)
      (error "a player's position doesn't match the field")))
  (for ([b balls])
    (unless (eq? (player-ball (ball-carried-by b)) b)
      (error "ball's carrier does not refer to ball"))
    (unless (and (= (ball-x b) (player-x (ball-carried-by b)))
                 (= (ball-y b) (player-y (ball-carried-by b))))
      (error "ball is not at its carrier's position"))))

(define-struct contest (task p1 p2 inc!))

;; serve : list-of-player list-of-contest -> void
;; Waits for
;;   - msg from any player in `players'
;;   - end of a contest in `contests'
;; When a contest starts, remove the two players in the contest from `players'
;; When a contest ends, remove from `contests', add the two players back to `players'
;; When handling a request with a `quit' response, remove the player from `players'
(define (serve players contests)
  (unless (and (null? players) 
               (null? contests))
    (apply sync 
           (append
            ;; Wait for a message from each player in `players':
            (for/list ([p players])
              (handle-evt
               (task-evt (player-task p))
               (lambda (req)
                 ;; Handle request and call `serve':
                 (handle-player p req players contests))))
            ;; Wait for a contest to end
            (for/list ([c contests])
              (handle-evt
               (task-evt (contest-task c))
               (lambda (transfer?)
                 ;; End contest and call `serve':
                 (handle-contest c transfer? players contests))))))))

(define (handle-player p req players contests)
  (define t (player-task p))
  (match req
    [(state-request)
     (task-put t (state (player-x p)
                        (player-y p)
                        (ball? (player-ball p))))
     (serve players contests)]
    [(move-request x y)
     (define p2 (field-ref x y))
     (cond
       [(= moves MAX-MOVES) 
        (task-put t (quit))
        (serve (remq p players) contests)]
       [(player? p2)
        (define same-team? (eq? (player-team p) (player-team p2)))
        (cond
          [(not (memq p2 players))
           (task-put t (bump))
           (serve players contests)]
          [(and same-team?
                (player-ball p)
                (not (player-ball p2)))
           (task-put t (pass))
           (inc-moves!)
           (serve (remq p (remq p2 players))
                  (cons (contest (player-task p) p p2 (make-inc-passes p))
                        contests))]
          [(and (not same-team?)
                (player-ball p2)
                (not (player-ball p)))
           (task-put t (steal))
           (inc-moves!)
           (serve (remq p (remq p2 players))
                  (cons (contest (player-task p) p2 p (make-inc-steals p))
                        contests))]
          [else
           (task-put t (bump))
           (serve players contests)])]
       [else
        (field-set! (player-x p) (player-y p) #f)
        (field-set! x y p)
        (set-player-x! p x)
        (set-player-y! p y)
        (define b (player-ball p))
        (when b
          (set-ball-x! b x)
          (set-ball-y! b y))
        (task-put t (move))
        (inc-moves!)
        (serve players contests)])]
    [else (error "bad message")]))

(define (handle-contest c transfer? players contests)
  (match c
    [(contest t p1 p2 inc!)
     (define b (player-ball p1))
     (when transfer?
       (inc!)
       (set-player-ball! p2 b)
       (set-player-ball! p1 #f)
       (set-ball-carried-by! b p2)
       (set-ball-x! b (player-x p2))
       (set-ball-y! b (player-y p2)))
     (serve (cons p1 (cons p2 players)) (remq c contests))]))

(define (make-inc-passes p)
  (if (eq? #\r (player-team p))
      (lambda () (set! red-passes (add1 red-passes)))
      (lambda () (set! blue-passes (add1 blue-passes)))))

(define (make-inc-steals p)
  (if (eq? #\r (player-team p))
      (lambda () (set! red-steals (add1 red-steals)))
      (lambda () (set! blue-steals (add1 blue-steals)))))

(define (inc-moves!)
  (set! moves (add1 moves))
  (when (zero? (modulo moves MOVES-PER-EPOCH))
    (show-field)))

(show-field)
(serve players null)
(show-field)
